from __future__ import annotations

import json
import re
from dataclasses import dataclass
from inspect import Parameter, getdoc, signature
from typing import Any, Callable

from pydantic import BaseModel, ConfigDict, Field, ValidationError, create_model


class ToolDefinition:
    """
    A tool available to the LLM.
    """

    name: str
    func: Callable
    model: type[BaseModel]
    args_schema: dict[str, Any]

    def __init__(self, func: Callable, name: str | None = None):
        """
        Define a new tool that runs a function.

        The function should have a Sphinx-style docstring with :param: and
        :return: lines to describe the parameters and return value. The
        docstring should describe the function in detail so that the LLM
        can decide when to use it and to know how.

        :param func: The function to run.
        :param name: The name of the tool, exposed to the LLM. Defaults to the
            function name.
        """
        self.name = name or func.__name__
        self.func = func
        self.model = self.create_model_from_function(func)

        self.schema = self.model.model_json_schema()
        self.schema.pop("title", None)
        self.description = self.schema.pop("description", None)

    @staticmethod
    def parse_docstring(docstring: str):
        """
        Parse the Sphinx-style docstring and extract parameter descriptions.
        """
        param_pattern = r":param (\w+): (.+)"
        param_descriptions = {}
        if docstring:
            matches = re.findall(param_pattern, docstring)
            for param, description in matches:
                param_descriptions[param] = description.strip()
        return param_descriptions

    @classmethod
    def create_model_from_function(cls, func):
        sig = signature(func)
        docstring = getdoc(func)
        param_descriptions = cls.parse_docstring(docstring)
        fields = {}

        for name, param in sig.parameters.items():
            annotation = (
                param.annotation if param.annotation != Parameter.empty else Any
            )
            default = param.default if param.default != Parameter.empty else ...
            desc = param_descriptions.get(name)
            if desc:
                fields[name] = (annotation, Field(default, description=desc))
            else:
                fields[name] = (annotation, default)

        model_name = (
            "".join(part.capitalize() for part in func.__name__.split("_")) + "Args"
        )

        model = create_model(
            model_name,
            __doc__=docstring,
            __config__=ConfigDict(extra="forbid"),
            **fields,
        )

        return model


@dataclass
class ToolCall:
    id: str
    name: str
    arguments: dict[str, Any]

    @classmethod
    def from_openai(cls, tool_call: Any):
        from openai.types.chat import ParsedFunctionToolCall

        assert isinstance(tool_call, ParsedFunctionToolCall)

        return cls(
            id=tool_call.id,
            name=tool_call.function.name,
            arguments=json.loads(tool_call.function.arguments),
        )

    @classmethod
    def from_anthropic(cls, block: Any):
        # assert isinstance(block, ToolUseBlock)

        return cls(
            id=block.id,
            name=block.name,
            arguments=json.dumps(block.input),
        )

    @property
    def json_message(self):
        return {
            "id": self.id,
            "type": "function",
            "function": {
                "name": self.name,
                "arguments": self.arguments,
            },
        }


@dataclass
class ToolResponse:
    call: ToolCall
    response: str = None
    error: str = None
    # exception also!


class ToolError(Exception):
    pass


class ToolKit:
    tools: dict[str, ToolDefinition]

    def __init__(self, functions: list[Callable]):
        tool_defs = [ToolDefinition(func) for func in functions]
        self.tools = {t.name: t for t in tool_defs}

    @property
    def tool_names(self) -> list[str]:
        return list(self.tools.keys())

    async def execute_tool_call(self, call: ToolCall) -> ToolResponse:
        tool = self.tools.get(call.name)
        if not tool:
            return ToolResponse(call=call, error=f"ERROR: Unknown tool: {call.name}")

        try:
            args = tool.model(**call.arguments)
        except (TypeError, ValidationError) as err:
            return ToolResponse(
                call=call,
                error=f"ERROR: Error parsing arguments for {call.name}: {err}",
            )

        try:
            response_text: str = tool.func(**args.__dict__)
            return ToolResponse(call=call, response=response_text)
        except ToolError as err:
            return ToolResponse(
                call=call, error=f"ERROR: Error running tool {call.name}: {err}"
            )

    def generate_tool_spec(
        self,
        formatter: Callable[[list[ToolDefinition]], dict],
    ) -> list[dict]:
        return [formatter(t) for t in self.tools.values()]



from unittest.mock import AsyncMock, MagicMock, call, patch

import pytest
from openai import NOT_GIVEN
from pydantic import BaseModel
from test_openai_adapter import (
    BASIC_CHAT,
    BASIC_OPENAI_MESSAGES,
)

from think.llm.chat import Chat
from think.llm.openai import OpenAIClient


def response_fixture(msg: dict):
    return MagicMock(
        choices=[
            MagicMock(
                message=MagicMock(
                    model_dump=MagicMock(return_value=msg),
                ),
            ),
        ]
    )


def parsed_fixture(msg: dict, val: BaseModel):
    fixture = response_fixture(msg)
    fixture.choices[0].message.parsed = val
    return fixture


@pytest.mark.asyncio
@patch("think.llm.openai.AsyncOpenAI")
async def test_call_minimal(AsyncOpenAI):
    chat = Chat.load(BASIC_CHAT)
    client = OpenAIClient(api_key="fake-key", model="fake-model")

    mock_create = AsyncMock(return_value=response_fixture(BASIC_OPENAI_MESSAGES[-1]))
    AsyncOpenAI.return_value.chat.completions.create = mock_create

    response = await client(chat)

    assert response == "Hi!"
    mock_create.assert_called_once_with(
        model="fake-model",
        messages=BASIC_OPENAI_MESSAGES,
        tools=NOT_GIVEN,
        temperature=None,
        max_completion_tokens=NOT_GIVEN,
    )

    AsyncOpenAI.assert_called_once_with(api_key="fake-key")


@pytest.mark.asyncio
@patch("think.llm.openai.AsyncOpenAI")
async def test_call_with_options(AsyncOpenAI):
    chat = Chat.load(BASIC_CHAT)
    client = OpenAIClient(api_key="fake-key", model="fake-model")

    mock_create = AsyncMock(return_value=response_fixture(BASIC_OPENAI_MESSAGES[-1]))
    AsyncOpenAI.return_value.chat.completions.create = mock_create

    response = await client(chat, temperature=0.5, max_tokens=10)

    assert response == "Hi!"
    mock_create.assert_called_once_with(
        model="fake-model",
        messages=BASIC_OPENAI_MESSAGES,
        tools=NOT_GIVEN,
        temperature=0.5,
        max_completion_tokens=10,
    )


@pytest.mark.asyncio
@patch("think.llm.openai.AsyncOpenAI")
async def test_call_with_tools(AsyncOpenAI):
    chat = Chat.load(BASIC_CHAT)
    client = OpenAIClient(api_key="fake-key", model="fake-model")

    tool_defs = [
        {
            "type": "function",
            "function": {
                "name": "fake_tool",
                "description": "Do something",
                "arguments": {
                    "additionalProperties": False,
                    "properties": {
                        "a": {"title": "A", "type": "integer"},
                        "b": {"title": "B", "type": "string"},
                    },
                    "required": ["a", "b"],
                    "type": "object",
                },
            },
        }
    ]

    tool_call = {
        "role": "assistant",
        "content": None,
        "tool_calls": [
            {
                "id": "call_boPTOC8z660AYdRFr9oogH4O",
                "type": "function",
                "function": {
                    "name": "fake_tool",
                    "arguments": '{"a": 1, "b": "hi"}',
                },
            }
        ],
    }

    mock_create = AsyncMock(return_value=response_fixture(tool_call))
    AsyncOpenAI.return_value.chat.completions.create = mock_create

    def fake_tool(a: int, b: str) -> str:
        """Do something"""
        assert a == 1
        assert b == "hi"
        return "tool response"

    response = await client(chat, tools=[fake_tool], max_steps=1)

    assert response == ""
    mock_create.assert_has_calls(
        [
            call(
                model="fake-model",
                messages=BASIC_OPENAI_MESSAGES,
                tools=tool_defs,
                temperature=None,
                max_completion_tokens=NOT_GIVEN,
            ),
            call(
                model="fake-model",
                messages=BASIC_OPENAI_MESSAGES
                + [
                    tool_call,
                    {
                        "role": "tool",
                        "tool_call_id": "call_boPTOC8z660AYdRFr9oogH4O",
                        "content": "tool response",
                    },
                ],
                tools=tool_defs,
                temperature=None,
                max_completion_tokens=NOT_GIVEN,
            ),
        ]
    )


@pytest.mark.asyncio
@patch("think.llm.openai.AsyncOpenAI")
async def test_call_with_pydantic(AsyncOpenAI):
    chat = Chat.load(BASIC_CHAT)
    client = OpenAIClient(api_key="fake-key", model="fake-model")

    class TestModel(BaseModel):
        text: str

    mock_parse = AsyncMock(
        return_value=parsed_fixture(
            BASIC_OPENAI_MESSAGES[-1],
            TestModel(text="Hi!"),
        )
    )
    AsyncOpenAI.return_value.beta.chat.completions.parse = mock_parse

    response = await client(chat, parser=TestModel)

    assert response.text == "Hi!"
    mock_parse.assert_called_once_with(
        model="fake-model",
        messages=BASIC_OPENAI_MESSAGES,
        tools=NOT_GIVEN,
        temperature=None,
        response_format=TestModel,
        max_completion_tokens=NOT_GIVEN,
    )

    AsyncOpenAI.assert_called_once_with(api_key="fake-key")


@pytest.mark.asyncio
@patch("think.llm.openai.AsyncOpenAI")
async def test_call_with_custom_parser(AsyncOpenAI):
    chat = Chat.load(BASIC_CHAT)
    client = OpenAIClient(api_key="fake-key", model="fake-model")

    mock_create = AsyncMock(return_value=response_fixture(BASIC_OPENAI_MESSAGES[-1]))
    AsyncOpenAI.return_value.chat.completions.create = mock_create

    def custom_parser(val: str) -> float:
        assert val == "Hi!"
        return 0.5

    response = await client(chat, parser=custom_parser)

    assert response == 0.5
    mock_create.assert_called_once_with(
        model="fake-model",
        messages=BASIC_OPENAI_MESSAGES,
        tools=NOT_GIVEN,
        temperature=None,
        max_completion_tokens=NOT_GIVEN,
    )

    AsyncOpenAI.assert_called_once_with(api_key="fake-key")



from unittest.mock import AsyncMock, MagicMock, call, patch

import pytest
from anthropic import NOT_GIVEN
from pydantic import BaseModel
from test_anthropic_adapter import (
    BASIC_ANTHROPIC_MESSAGES,
    BASIC_CHAT,
)

from think.llm.anthropic import AnthropicClient
from think.llm.chat import Chat


def response_fixture(msg: dict):
    return MagicMock(model_dump=MagicMock(return_value=msg))


@pytest.mark.asyncio
@patch("think.llm.anthropic.AsyncAnthropic")
async def test_call_minimal(AsyncAnthropic):
    chat = Chat.load(BASIC_CHAT)
    client = AnthropicClient(api_key="fake-key", model="fake-model")

    mock_create = AsyncMock(return_value=response_fixture(BASIC_ANTHROPIC_MESSAGES[-1]))
    AsyncAnthropic.return_value.messages.create = mock_create

    response = await client(chat)

    assert response == "Hi!"
    mock_create.assert_called_once_with(
        model="fake-model",
        messages=BASIC_ANTHROPIC_MESSAGES,
        tools=NOT_GIVEN,
        temperature=NOT_GIVEN,
        max_tokens=4096,
    )

    AsyncAnthropic.assert_called_once_with(api_key="fake-key")


@pytest.mark.asyncio
@patch("think.llm.anthropic.AsyncAnthropic")
async def test_call_with_options(AsyncAnthropic):
    chat = Chat.load(BASIC_CHAT)
    client = AnthropicClient(api_key="fake-key", model="fake-model")

    mock_create = AsyncMock(return_value=response_fixture(BASIC_ANTHROPIC_MESSAGES[-1]))
    AsyncAnthropic.return_value.messages.create = mock_create

    response = await client(chat, temperature=0.5, max_tokens=10)

    assert response == "Hi!"
    mock_create.assert_called_once_with(
        model="fake-model",
        messages=BASIC_ANTHROPIC_MESSAGES,
        tools=NOT_GIVEN,
        temperature=0.5,
        max_tokens=10,
    )


@pytest.mark.asyncio
@patch("think.llm.anthropic.AsyncAnthropic")
async def test_call_with_tools(AsyncAnthropic):
    chat = Chat.load(BASIC_CHAT)
    client = AnthropicClient(api_key="fake-key", model="fake-model")

    tool_call = {
        "role": "assistant",
        "content": [
            {
                "type": "tool_use",
                "id": "call_boPTOC8z660AYdRFr9oogH4O",
                "name": "fake_tool",
                "input": {"a": 1, "b": "hi"},
            }
        ],
    }

    tool_defs = [
        {
            "name": "fake_tool",
            "description": "Do something",
            "input_schema": {
                "additionalProperties": False,
                "properties": {
                    "a": {"title": "A", "type": "integer"},
                    "b": {"title": "B", "type": "string"},
                },
                "required": ["a", "b"],
                "type": "object",
            },
        }
    ]

    mock_create = AsyncMock(return_value=response_fixture(tool_call))
    AsyncAnthropic.return_value.messages.create = mock_create

    def fake_tool(a: int, b: str) -> str:
        """Do something"""
        assert a == 1
        assert b == "hi"
        return "tool response"

    response = await client(chat, tools=[fake_tool], max_steps=1)

    assert response == ""
    mock_create.assert_has_calls(
        [
            call(
                model="fake-model",
                messages=BASIC_ANTHROPIC_MESSAGES,
                temperature=NOT_GIVEN,
                tools=tool_defs,
                max_tokens=4096,
            ),
            call(
                model="fake-model",
                messages=BASIC_ANTHROPIC_MESSAGES
                + [
                    tool_call,
                    {
                        "role": "user",
                        "content": [
                            {
                                "type": "tool_result",
                                "tool_use_id": "call_boPTOC8z660AYdRFr9oogH4O",
                                "content": "tool response",
                            },
                        ],
                    },
                ],
                temperature=NOT_GIVEN,
                tools=tool_defs,
                max_tokens=4096,
            ),
        ]
    )


@pytest.mark.asyncio
@patch("think.llm.anthropic.AsyncAnthropic")
async def test_call_with_pydantic(AsyncAnthropic):
    chat = Chat.load(BASIC_CHAT)
    client = AnthropicClient(api_key="fake-key", model="fake-model")

    class TestModel(BaseModel):
        text: str

    mock_create = AsyncMock(
        return_value=response_fixture(
            {"role": "assistant", "content": '{"text": "Hi!"}'}
        )
    )
    AsyncAnthropic.return_value.messages.create = mock_create

    response = await client(chat, parser=TestModel)

    assert response.text == "Hi!"
    mock_create.assert_called_once_with(
        model="fake-model",
        messages=BASIC_ANTHROPIC_MESSAGES,
        tools=NOT_GIVEN,
        temperature=NOT_GIVEN,
        max_tokens=4096,
    )

    AsyncAnthropic.assert_called_once_with(api_key="fake-key")


@pytest.mark.asyncio
@patch("think.llm.anthropic.AsyncAnthropic")
async def test_call_with_custom_parser(AsyncAnthropic):
    chat = Chat.load(BASIC_CHAT)
    client = AnthropicClient(api_key="fake-key", model="fake-model")

    mock_create = AsyncMock(return_value=response_fixture(BASIC_ANTHROPIC_MESSAGES[-1]))
    AsyncAnthropic.return_value.messages.create = mock_create

    def custom_parser(val: str) -> float:
        assert val == "Hi!"
        return 0.5

    response = await client(chat, parser=custom_parser)

    assert response == 0.5
    mock_create.assert_called_once_with(
        model="fake-model",
        messages=BASIC_ANTHROPIC_MESSAGES,
        tools=NOT_GIVEN,
        temperature=NOT_GIVEN,
        max_tokens=4096,
    )

    AsyncAnthropic.assert_called_once_with(api_key="fake-key")



from __future__ import annotations

from base64 import b64decode
from json import JSONDecodeError, loads
from logging import getLogger
from time import time
from typing import AsyncGenerator, Callable, Literal

try:
    from anthropic import NOT_GIVEN, AsyncAnthropic, AsyncStream
    from anthropic.types import Message as AnthropicMessage
    from anthropic.types import RawMessageStreamEvent
    from anthropic.types.image_block_param import Source
except ImportError as err:
    raise ImportError(
        "Anthropic client requires the Anthropic Python SDK: pip install anthropic"
    ) from err

from pydantic import BaseModel, ValidationError

from .base import LLM, CustomParserResultT, PydanticResultT
from .chat import Chat, ContentPart, ContentType, Message, Role
from .tool import ToolCall, ToolDefinition, ToolKit, ToolResponse

log = getLogger(__name__)


class AnthropicAdapter:
    toolkit: ToolKit

    def __init__(self, toolkit: ToolKit | None = None):
        self.toolkit = toolkit

    def get_tool_spec(self, tool: ToolDefinition) -> dict:
        return {
            "name": tool.name,
            "description": tool.description,
            "input_schema": tool.schema,
        }

    @property
    def spec(self) -> list[dict] | None:
        if self.toolkit is None:
            return NOT_GIVEN

        return self.toolkit.generate_tool_spec(self.get_tool_spec)

    def dump_role(self, role: Role) -> Literal["user", "assistant"]:
        if role in [Role.system, Role.user, Role.tool]:
            return "user"
        else:
            return "assistant"

    def dump_content_part(self, part: ContentPart) -> dict:
        match part:
            case ContentPart(type=ContentType.text, text=text):
                return dict(
                    type="text",
                    text=text,
                )
            case ContentPart(type=ContentType.image):
                return dict(
                    type="image",
                    source=Source(
                        type="base64",
                        data=part.image_data,
                        media_type=part.image_mime_type,
                    ),
                )
            case ContentPart(
                type=ContentType.tool_call,
                tool_call=ToolCall(id=id, name=name, arguments=arguments),
            ):
                return dict(
                    type="tool_use",
                    id=id,
                    name=name,
                    input=arguments,
                )
            case ContentPart(
                type=ContentType.tool_response,
                tool_response=ToolResponse(
                    call=ToolCall(id=id),
                    response=response,
                ),
            ):
                return dict(
                    type="tool_result",
                    tool_use_id=id,
                    content=response,
                )
            case _:
                raise ValueError(f"Unknown content type: {part.type}")

    def parse_content_part(self, part: dict) -> ContentPart:
        match part:
            case {"type": "text", "text": text}:
                return ContentPart(type=ContentType.text, text=text)
            case {"type": "image", "source": {"data": data}}:
                return ContentPart(
                    type=ContentType.image,
                    image=b64decode(data.encode("ascii")),
                )
            case {"type": "tool_use", "id": id, "name": name, "input": input}:
                return ContentPart(
                    type=ContentType.tool_call,
                    tool_call=ToolCall(id=id, name=name, arguments=input),
                )
            case {"type": "tool_result", "tool_use_id": id, "content": content}:
                return ContentPart(
                    type=ContentType.tool_response,
                    tool_response=ToolResponse(
                        call=ToolCall(id=id, name="", arguments={}),
                        response=content,
                    ),
                )
            case _:
                raise ValueError(f"Unknown content type: {part.type}")

    def dump_message(self, message: Message) -> dict:
        if len(message.content) == 1 and message.content[0].type == ContentType.text:
            content = message.content[0].text
        else:
            content = [self.dump_content_part(part) for part in message.content]

        return dict(
            role=self.dump_role(message.role),
            content=content,
        )

    def parse_message(self, message: dict | AnthropicMessage) -> Message:
        if isinstance(message, AnthropicMessage):
            message = message.model_dump()

        role = Role.assistant if message.get("role") == "assistant" else Role.user
        content = message.get("content")
        if isinstance(content, str):
            return Message(
                role=role,
                content=[
                    ContentPart(type=ContentType.text, text=content),
                ],
            )

        parts = [self.parse_content_part(part) for part in content]
        if any(part.type == ContentType.tool_response for part in parts):
            role = Role.tool
        return Message(role=role, content=parts)

    def dump_chat(self, chat: Chat) -> tuple[str, list[dict]]:
        system_messages = []
        other_messages = []
        offset = 0

        # If the first message is a system one, extract it as a separate
        # string argument, but *only* if there are more messages. Otherwise
        # include it as usual (note that "system" role will be automatically
        # converted to "user").
        if len(chat) > 1 and chat.messages[0].role == Role.system:
            for part in chat.messages[0].content:
                system_messages.append(part.text)
            offset = 1

        for msg in chat.messages[offset:]:
            other_messages.append(self.dump_message(msg))

        system_message = "\n\n".join(system_messages) if system_messages else NOT_GIVEN
        return system_message, other_messages

    def load_chat(self, messages: list[dict], system: str | None = None) -> Chat:
        c = Chat(system)
        for m in messages:
            c.messages.append(self.parse_message(m))
        return c


class AnthropicClient(LLM):
    provider = "anthropic"

    def __init__(
        self,
        model: str,
        *,
        api_key: str | None = None,
        base_url: str | None = None,
    ):
        super().__init__(model, api_key=api_key, base_url=base_url)
        self.client = AsyncAnthropic(api_key=api_key)

    async def __call__(
        self,
        chat: Chat,
        *,
        parser: type[PydanticResultT]
        | Callable[[str], CustomParserResultT]
        | None = None,
        temperature: float | None = None,
        tools: ToolKit | list[Callable] | None = None,
        max_retries: int = 3,
        max_steps: int = 5,
        max_tokens: int | None = None,
    ) -> str | PydanticResultT | CustomParserResultT:
        toolkit = self._get_toolkit(tools)

        if max_tokens is None:
            max_tokens = 4096

        adapter = AnthropicAdapter(toolkit)
        system_message, messages = adapter.dump_chat(chat)
        tools_dbg = f" and tools {{{toolkit.tool_names}}}" if toolkit else ""
        log.debug(f"Making a {self.model} call with messages: {messages}{tools_dbg}")
        t0 = time()
        anthropic_message: AnthropicMessage = await self.client.messages.create(
            model=self.model,
            messages=messages,
            temperature=NOT_GIVEN if temperature is None else temperature,
            tools=adapter.spec,
            max_tokens=max_tokens,
        )
        t1 = time()

        anthropic_message = anthropic_message.model_dump()
        log.debug(f"Received response in {(t1 - t0):.1f}s: {anthropic_message}")

        message = adapter.parse_message(anthropic_message)
        text, response_list = await self._process_message(chat, message, toolkit)

        if response_list:
            if max_steps < 1:
                log.warning("Tool call steps limit reached, stopping")
            else:
                return await self(
                    chat,
                    temperature=temperature,
                    tools=tools,
                    parser=parser,
                    max_steps=max_steps - 1,
                    max_retries=max_retries,
                )

        if parser:
            try:
                if isinstance(parser, type) and issubclass(parser, BaseModel):
                    return parser(**loads(text))
                else:
                    return parser(text)

            except (JSONDecodeError, ValidationError, ValueError) as err:
                log.debug(f"Error parsing response '{text}': {err}")

                if not max_retries:
                    raise

                error_text = f"Error parsing your response: {err}. Please output your response EXACTLY as requested."
                chat.user(error_text)
                return await self(
                    chat,
                    temperature=temperature,
                    tools=tools,
                    parser=parser,
                    max_steps=max_steps,
                    max_retries=max_retries - 1,
                )

        return text

    async def stream(
        self,
        chat: Chat,
        temperature: float | None = None,
        max_tokens: int | None = None,
    ) -> AsyncGenerator[str, None]:
        if max_tokens is None:
            max_tokens = 4096

        adapter = AnthropicAdapter()
        system_message, messages = adapter.dump_chat(chat)

        log.debug(f"Making a {self.model} stream call with messages: {messages}")
        stream: AsyncStream[RawMessageStreamEvent] = await self.client.messages.create(
            model=self.model,
            messages=messages,
            temperature=NOT_GIVEN if temperature is None else temperature,
            stream=True,
            system=system_message,
            max_tokens=max_tokens,
        )
        text = ""
        async for event in stream:
            if event.type == "content_block_delta":
                text += event.delta.text
                yield event.delta.text
            else:
                log.debug("AnthropicClient.stream(): ignoring unknown chunk %r", event)

        if text:
            chat.messages.append(
                Message(
                    role=Role.assistant,
                    content=[ContentPart(type=ContentType.text, text=text)],
                )
            )



from pathlib import Path
from typing import Any, Optional

from jinja2 import BaseLoader, Environment, FileSystemLoader, StrictUndefined


def strip_block(txt: str) -> str:
    """
    Strip a multiline block

    Strips whitespace from each line in the block so that the indentation
    (if any) within the block is preserved, but the block itself is not
    indented. Also strips any trailing whitespace.

    :param txt: The block of text to strip.
    :return: The stripped block of text.
    """
    lines = txt.splitlines()
    min_indent = min(len(line) - len(line.lstrip()) for line in lines if line.strip())
    return "\n".join(line[min_indent:].rstrip() for line in lines).strip("\n")


class FormatTemplate:
    def __call__(self, template: str, **kwargs: dict[str, Any]) -> str:
        return strip_block(template).format(**kwargs)


class BaseJinjaTemplate:
    def __init__(self, loader: Optional[BaseLoader]):
        self.env = Environment(
            loader=loader,
            autoescape=False,
            lstrip_blocks=True,
            trim_blocks=True,
            keep_trailing_newline=False,
            undefined=StrictUndefined,
        )


class JinjaStringTemplate(BaseJinjaTemplate):
    def __init__(self):
        super().__init__(None)

    def __call__(self, template: str, **kwargs: dict[str, Any]) -> str:
        tpl = self.env.from_string(strip_block(template))
        return tpl.render(**kwargs)


class JinjaFileTemplate(BaseJinjaTemplate):
    def __init__(self, template_dir: str):
        if not Path(template_dir).is_dir():
            raise ValueError(f"Template directory does not exist: {template_dir}")
        super().__init__(FileSystemLoader(template_dir))

    def __call__(self, template: str, **kwargs: dict[str, Any]) -> str:
        tpl = self.env.get_template(template)
        return tpl.render(**kwargs)



from __future__ import annotations

from abc import ABC, abstractmethod
from logging import getLogger
from typing import AsyncGenerator, Callable, TypeVar, overload
from urllib.parse import urlparse

from pydantic import BaseModel

from .chat import Chat, ContentPart, ContentType, Message, Role
from .tool import ToolKit, ToolResponse

CustomParserResultT = TypeVar("CustomParserResultT")
PydanticResultT = TypeVar("PydanticResultT", bound=BaseModel)

log = getLogger(__name__)


class LLM(ABC):
    PROVIDERS = ["openai", "anthropic"]
    provider: str
    base_url: str | None = None
    model: str

    def __init__(
        self,
        model: str,
        *,
        api_key: str | None = None,
        base_url: str | None = None,
    ):
        self.model = model
        self.api_key = api_key
        self.base_url = base_url

    @classmethod
    def for_provider(cls, provider: str) -> type["LLM"]:
        if provider == "openai":
            from .openai import OpenAIClient

            return OpenAIClient
        elif provider == "anthropic":
            from .anthropic import AnthropicClient

            return AnthropicClient

        else:
            raise ValueError(f"Unknown provider: {provider}")

    @classmethod
    def from_url(cls, url: str) -> "LLM":
        result = urlparse(url)
        return cls.for_provider(result.scheme)(
            model=result.path.lstrip("/"),
            api_key=result.username,
            base_url=result.hostname,
        )

    def _get_toolkit(
        self,
        tools: ToolKit | list[Callable[[str], str]] | None,
    ) -> ToolKit | None:
        if tools is None:
            return None

        if isinstance(tools, ToolKit):
            return tools

        return ToolKit(tools)

    @overload
    @abstractmethod
    async def __call__(
        self,
        chat: Chat,
        *,
        parser: type[PydanticResultT],
        temperature: float | None = None,
        tools: ToolKit | list[Callable] | None = None,
        max_retries: int = 3,
        max_steps: int = 5,
        max_tokens: int | None = None,
    ) -> PydanticResultT: ...

    @overload
    @abstractmethod
    async def __call__(
        self,
        chat: Chat,
        *,
        parser: Callable[[str], CustomParserResultT],
        temperature: float | None = None,
        tools: ToolKit | list[Callable] | None = None,
        max_retries: int = 3,
        max_steps: int = 5,
        max_tokens: int | None = None,
    ) -> CustomParserResultT: ...

    @overload
    @abstractmethod
    async def __call__(
        self,
        chat: Chat,
        *,
        temperature: float | None = None,
        tools: ToolKit | list[Callable] | None = None,
        max_retries: int = 3,
        max_steps: int = 5,
        max_tokens: int | None = None,
    ) -> str: ...

    @abstractmethod
    async def __call__(
        self,
        chat: Chat,
        *,
        parser: type[PydanticResultT]
        | Callable[[str], CustomParserResultT]
        | None = None,
        temperature: float | None = None,
        tools: ToolKit | list[Callable] | None = None,
        max_retries: int = 3,
        max_steps: int = 5,
        max_tokens: int | None = None,
    ) -> str | PydanticResultT | CustomParserResultT:
        """Make an LLM call with the chat conversation and return a response.

        :param chat: The chat conversation to process
        :param parser: Optional parser to process the response. Can be either:
            - A Pydantic model class to parse the response into
            - A callable that takes a string and returns a custom type
        :param temperature: Optional sampling temperature (0-1)
        :param tools: Optional tools/functions available to the model:
            - A ToolKit instance
            - A list of callables that take a string and return a string
        :param max_retries: Maximum number of retries on failure
        :param max_steps: Maximum number of steps for tool use
        :param max_tokens: Optional maximum tokens in response
        :return: Either:
            - An instance of the provided Pydantic model if a model class was provided as parser
            - The result of the parser callable if a custom parser was provided
            - The raw string response if no parser was provided
        :raises ValueError: If the temperature is not between 0 and 1
        :raises APIError: If there is an error communicating with the API
        """
        ...

    async def _process_message(
        self,
        chat: Chat,
        message: Message,
        toolkit: ToolKit,
    ) -> tuple[str, list[ToolResponse]]:
        chat.messages.append(message)

        text = ""
        response_list = []
        for part in message.content:
            if part.type == ContentType.text:
                text += part.text
            elif part.type == ContentType.tool_call:
                if toolkit is None:
                    log.warning("Tool call with no toolkit defined, ignoring")
                    continue
                tool_response = await toolkit.execute_tool_call(part.tool_call)
                response_list.append(tool_response)

        if response_list:
            chat.messages.append(
                Message(
                    role=Role.tool,
                    content=[
                        ContentPart(
                            type=ContentType.tool_response,
                            tool_response=tool_response,
                        )
                        for tool_response in response_list
                    ],
                )
            )

        return text, response_list

    @abstractmethod
    async def stream(
        self,
        chat: Chat,
        temperature: float | None = None,
        max_tokens: int | None = None,
    ) -> AsyncGenerator[str, None]:
        """Stream responses from the LLM model.

        :param chat: The chat conversation to process
        :param temperature: Optional sampling temperature (0-1)
        :param max_tokens: Optional maximum tokens in response
        :return: An async generator of response strings
        """
        ...
